/*
 * Copyright (C) 2012, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright (C) 2012, Jonathan Vacher, ENS Cachan <jvacher@ens-cachan.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>
#include <fftw3.h>
#include "periodic_component.h"
#include "toolbox.h"
#include "Eigen/Dense"
#include <vector>
#include "Eigen/Polynomials"

using namespace Eigen;
using namespace std;


/** print a message and abort the execution */
#define FATAL(MSG)			\
do {				\
	fprintf(stderr, MSG "\n");	\
	abort();			\
	} while (0);
	
/**
* @file parameters.c
* @brief Compute the bilinear zoom of an image.
*
* @version 0.1
* @author Thibaud BRIAND & Jonathan VACHER ; <thibaud.briand@ens-cachan.fr> ;
* <jvacher@ens-cachan.fr>
*
*/

/* moment part */

/**
* @brief Compute the mean of a 2D array of float type.
*
* @param data_in input array
* @param N size of the array
* @return mean of data_in
*/
float mean(float *data_in, size_t N)
{
	float m = 0;
	float *ptr, *ptr_end;

	/* Pointers initialization */
	ptr = data_in;
	ptr_end = ptr + N;

	/* Sum loop */
	while(ptr < ptr_end)
	{
		m += *ptr;
		ptr++;
	}

	/* Normalization */
	m /= ((float) N);

	return(m);
}

/**
* @brief Compute the orderth moment of an array of float type.
*
* @param data_in input array
* @param order of the moment
* @param N size of the array
* @return the orderth moment of data_in
*/
float compute_moment(float *data_in, int order, size_t N)
{
	
	float moment = 0;
	float m = mean(data_in, N);
	float *ptr, *ptr_end;
        float p;
	
	/* Pointers initialization */
	ptr = data_in;
	ptr_end = ptr + N;

	/* Sum loop */
	while(ptr < ptr_end)
	{
	  p=1;
	  for(int i=1;i<(order+1);i++) p *= (*ptr-m);

	  moment += p;
	  ptr++;
	}

	/* Normalization */
	moment /= ((float) N);

	return(moment);
}

/**
* @brief Compute the skewness of an array of float type.
*
* @param data_in input array
* @param N size of the array
* @return skewness of data_in
*/
float compute_skewness(float *data_in, size_t N)
{
	float sk;
	float order_2 = compute_moment(data_in, 2 , N);
	float order_3 = compute_moment(data_in, 3 , N);

  /* odd case */
  if(order_2==0)
    FATAL("Error in compute_skewness : you can not divide by 0");
	
	float sqrt_order_2_pow_3 = sqrt(order_2*order_2*order_2);
	
	sk = order_3/sqrt_order_2_pow_3;
	
	return(sk);
}

/**
* @brief Compute the kurtosis of an array of float type.
*
* @param data_in input array
* @param N size of the array
* @return kurtosis of data_in
*/
float compute_kurtosis(float *data_in, size_t N)
{
	float kur;
	float order_2 = compute_moment(data_in, 2 , N);
	float order_4 = compute_moment(data_in, 4 , N);

  /* odd case */
  if(order_2==0)
    FATAL("Error in compute_kurtosis : you can not divide by 0");
	
	kur = order_4/(order_2*order_2);

	return(kur);
}

/* range part */

/**
 * @brief Compute the minimum of a list of floats.
 *
 * @param tab input array
 * @param N size of the array
 * @return min of tab
 */
float min(float *tab, size_t N)
{
  float min=tab[0];
  for(unsigned int i=1;i<N;++i)
  {
     if (isnan(tab[i])) FATAL("Nan in min function");
     if (tab[i]<min) min=tab[i];
  }
  return(min);
}

/**
 * @brief Compute the maximum of a list of floats.
 *
 * @param tab input array
 * @param N size of the array
 * @return max of tab
 */
float max(float *tab, size_t N)
{
  float max=tab[0];
  for(unsigned int i=1;i<N;++i)
  {
    if (isnan(tab[i])) FATAL("Nan in max function");
     if (tab[i]>max) max=tab[i];
  }
  return(max);
}

/**
* @brief Ajust the range of an array
*
* @param data modified array
* @param N size of the array
* @param m,M desired range
*/
void mod_range(float *data, float m, float M, size_t N)
{
  /* allocation */
  int p=0;
  int q=0;

  /* loop on the array */
  for(unsigned int i=0;i<N;i++)
  {
    p = (data[i]<=m);
    q = (data[i]>=M);

    /* projection in [m,M] */
    data[i] = (float) p*m + q*M + (1-p)*(1-q)*data[i];
  }
}

/* Modification part of marginal statistics */

/**
* @brief Ajust the skewness of an array
*
* @param data float array
* @param sk_out desired skewness
* @param N size of the array
* @return data with modified skewness
* @note Note that the mean of data is still 0 after aplying mod_skewness
*/
void mod_skewness(float *data, float sk_out, size_t N)
{
/* compute the skewness of the input */
float sk_in = compute_skewness(data,N);
float m2,m4,m5,m6;
/* Avoid the odd case of similar skewness
 * don't modify the input
 */
	/* substract the mean */
	float m = mean(data,N);
	for(unsigned int i=0; i<N ; i++ ) data[i]-=m;

	/* memory allocation */
	std::vector<double> roots;
	double relevant_roots[6]={0};
	float final_roots[6]={0};
	double p0,p1,p2,p3,q0,q1,q2,a0,a1,a2,a3,a4,a5,a6;
	float lambda = 0;
	float g; // to compute the gradient

	  /* computation of the moments */
m2 = compute_moment(data,2,N);
m4 = compute_moment(data,4,N);
m5 = compute_moment(data,5,N);
m6 = compute_moment(data,6,N);

	/* define the coefficients of the numerator P */
	p0 = sk_in*m2*sqrt(m2);
	p1 = 3 * ( m4 - m2 * m2 * ( 1 + sk_in*sk_in ) );
	p2 = 3 * ( m5 - 2*sqrt(m2)*sk_in*m4 + m2*m2*sqrt(m2)*sk_in*sk_in*sk_in );
	p3 = m6 -3*sqrt(m2)*sk_in*m5 + 3*m2*(sk_in*sk_in - 1)*m4+m2*m2*m2*(2+3*sk_in*sk_in-sk_in*sk_in*sk_in*sk_in);

	/* define the coefficients of the denominator Q */
	q0 = m2;
	q1 = 0;
	q2 = m4 - (1+sk_in*sk_in)*m2*m2;

	/* define the polynomial A = P^2 - sk_out^2*Q^2 */
	a0 = p0*p0-sk_out*sk_out*q0*q0*q0;
	a1 = 2*p1*p0;
	a2 = p1*p1 + 2*p2*p0 - 3*sk_out*sk_out*q2*q0*q0;
	a3 = 2*(p3*p0 + p1*p2);
	a4 = p2*p2 + 2*p3*p1 - 3*sk_out*sk_out*q2*q2*q0;
	a5 = 2*p3*p2;
	a6 = p3*p3 - sk_out*sk_out*q2*q2*q2;

	VectorXd poly_num(4);
	poly_num << p0,p1,p2,p3;
	Vector3d poly_denom;
	poly_denom << q0,q1,q2;
	VectorXd poly(7);
	poly << a0,a1,a2,a3,a4,a5,a6;

 	PolynomialSolver< double, 6 > psolve( poly );
 	psolve.realRoots(roots);

	/* Keep the real roots (lambda is real)
	 * and with good sign
	 * sign(lambda) has to be sign(sk_out-sk_in)
	 */
	int sign0 = (sk_out>sk_in) - (sk_out<sk_in);
	int sign=0;
	int p=0; // number of roots that verify the two first criteria
	int q=0; // number of roots that verify the symmetric criterion
	int r=roots.size();

	for(int i=0;i<r;i++){
		sign = (int) (roots[i]>0) -(roots[i]<0);
		if(sign == sign0 && fabs(roots[i])<1e6)
		{
		  relevant_roots[p]= roots[i];
		  p++;
		}
	}

	/* odd case sk_out=0
	 * there is no symmetric condition */
	if((p>0) && (sk_out ==0)) //in fact p>0 if sk_out = 0
	{
	  for(int i =0;i<p;i++) final_roots[i]= (float) relevant_roots[i];
	  if(sign0>0) lambda = min( final_roots,p);
	  if(sign0<0) lambda = max( final_roots,p);
	  q=1; // a final solution has been found
	}

	/* Don't take the symmetric solution  i.e
	 * the sign of the numerator has to be the same as sk_out
	 * sign(P) = sign(sk_out)
	 */
	if((p>0) && (sk_out != 0)) // if there exists real solution and the skewness is not 0
	{
	  double numerator =0;
	  sign0 = (int) (sk_out>0)-(sk_out<0);

	  for(int i=0; i<p;i++)
	  {
	   numerator = poly_eval(poly_num, relevant_roots[i]);
	   sign = (int) (numerator>0)-(numerator<0);
	   if(sign==sign0)
	   {
	   final_roots[q]=(float) relevant_roots[i];
	   q++;
	   }
	  }

	  if(q>0) // if there exists a solution
	  {
	   if(sk_out>0) lambda = min(  final_roots,q);
	   if(sk_out<0) lambda = max(  final_roots,q);
	  }
	}

	/* define the translated array */
	  for(unsigned int i=0; i<N ; i++)
	  {
	  g = data[i]*(data[i]-sqrt(m2)*sk_in)-m2;
	  data[i]+= lambda*g;
	  }
	/* add the mean */
	for(unsigned int i=0; i<N ; i++) data[i]+=m;
 }

/**
* @brief Ajust the kurtosis of an array
*
* @param data float array
* @param ku_out desired kurtosis
* @param N size of the array
* @return data with modified kurtosis
* @note Note that the mean of data is still 0 after aplying mod_kurtosis
*/
void mod_kurtosis(float *data, float ku_out, size_t N)
{
  /* compute the skewness of the input */
  float ku_in = compute_kurtosis(data,N);

  /* memory allocation */
  std::vector<double> roots;
  double relevant_roots[4]={0};
  float final_roots[4]={0};
  float lambda = 0;
  float m1,m2,m3,m4,m5,m6,m7,m8,m9,m10,m12,alpha;
  double p0,p1,p2,p3,p4,q0,q1,q2;
  double a0,a1,a2,a3,a4;
  float g;

    /* computation of the moments */
  m1 = mean(data,N);
  m2 = compute_moment(data,2,N);
  m3 = compute_moment(data,3,N);
  m4 = compute_moment(data,4,N);
  m5 = compute_moment(data,5,N);
  m6 = compute_moment(data,6,N);
  m7 = compute_moment(data,7,N);
  m8 = compute_moment(data,8,N);
  m9 = compute_moment(data,9,N);
  m10 = compute_moment(data,10,N);
  m12 = compute_moment(data,12,N);

  alpha = m4/m2;

  for(unsigned int i=0;i<N;i++)
  {
    data[i] -= m1;
  }

  /* define the coefficients of the numerator */
  p0 = m4;
  p1 = 4 * ( m6 - alpha * alpha * m2 - m3 * m3 );
  p2 = 6 * ( m8 - 2 * alpha * m6 - 2 * m3 * m5 + alpha * alpha * m4 + (m2 + 2
  *alpha) * m3 * m3 );
  p3 = 4 * ( m10 - 3 * alpha * m8 - 3 * m3 * m7 + 3 * alpha * alpha * m6 + 6 *
  alpha * m3 * m5 + 3 * m3 * m3 * m4 - alpha * alpha * alpha * m4 - 3 * alpha *
  alpha * m3 * m3 - 3 * m4 * m3 * m3 );
  p4 = ( m12 - 4 * alpha * m10 - 4 * m3 * m9 + 6 * alpha * alpha * m8 + 12 * alpha
  * m3 * m7 +  6 * m3 * m3 * m6 - 4 * alpha * alpha * alpha * m6 - 12 * alpha *
  alpha * m3 * m5 + alpha * alpha * alpha * alpha * m4 - 12 * alpha * m3 * m3 * m4
  + 4 * alpha * alpha * alpha * m3 * m3 + 6 * alpha * alpha * m3 * m3 * m2
  - 3 * m3 * m3 * m3 * m3 );

  /* define the coefficients of the denominator */
  q0 = m2;
  q1 = 0;
  q2 = p1/4;

  /* define the polynomial */
  a0 = p0 - ku_out * q0 * q0;
  a1 = p1;
  a2 = p2 - 2 * ku_out * q2 * q0;
  a3 = p3;
  a4 = p4 - ku_out * q2 * q2;

  Vector3d poly_num;
  poly_num << q0,q1,q2;
  VectorXd poly_denom(5);
  poly_denom << p0,p1,p2,p3,p4;
  VectorXd poly(5);
  poly << a0,a1,a2,a3,a4;

   PolynomialSolver< double, 4 > psolve( poly );
   psolve.realRoots( roots );

  /* Keep the roots with good sign,
  * sign(lambda) has to be sign(ku_out-ku_in),
  * and which make the denominator non-zero
  */
  int p=0; // number of roots that verify the first criteria
  int q=0; // number of roots that verify the symmetric criterion
  int r=roots.size();

  for(int i=0;i<r;i++)
  {
    if( fabs(roots[i])<10e6)
    {
      relevant_roots[p]=roots[i];
      p++;
    }
  }

  /* The denominator must be non zero
  */
  if(p>0) // if there exists real solution verifying sign=sign0
  {
    double denominator = 0;

    for(int i=0; i<p;i++)
    {
    denominator = poly_eval(poly_denom,relevant_roots[i]);
    if(denominator>0)
    {
      final_roots[q]= (float) relevant_roots[i];
      q++;
    }
    }

    if(q>0) // if there exists a solution
    {
      lambda = final_roots[0];
      for(int i=1; i<q;i++)
      {
	if(fabs(final_roots[i])<fabs(lambda))
	{
	  lambda = final_roots[i];
	}
      }
    }
  }

  /* define the translated array
   * avoid the odd case of similar skewness */
 float test = log10(ku_in*ku_in/((ku_out-ku_in)*(ku_out-ku_in)));
 if(test>6)
 {
  for(unsigned int i=0; i<N ; i++)
  {
    g = data[i] * (data[i] * data[i] - alpha ) - m3;
    data[i] += lambda * g+m1;
  }
 }
}
/**
* @brief Ajust the mean and the variance of an array
*
* @param data modified array
* @param N size of the array
* @param mean_out desired mean
* @param var_out desired variance
*/
void mod_mean_variance(float *data, float mean_out, float var_out, size_t N)
{
  /* allocation of the input variance */
  float m_in = mean(data,N);
  float var_in = compute_moment(data,2,N);

  /* odd case */
  if(var_in==0)
    FATAL("Error in mod_mean_variance : you can not divide by 0");

  /* loop on the array */
  for(unsigned int i=0;i<N;i++)
  {
    data[i] = sqrt(var_out/var_in)*(data[i]-m_in)+mean_out;
  }
}

/* Correlation part */

/**
* @brief Compute the auto correlation of array of float type.
*
* @param list_data input arrays
* @param Na size of the neighbourhood
* @param nx,ny sizes of the array
* @return Auto correlation matrix of list_data (double matrix of size Na*Na)
*/
MatrixXd compute_auto_cor(float *list_data, size_t nx, size_t ny, size_t Na)
{
  fftwf_complex *fft_list_data = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));
  fftwf_complex *squarred_fft_list_data = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));
  float *ifft_list_data = (float *) malloc(nx*ny*sizeof(float));
  float *list_data_0 = (float *) malloc(nx*ny*sizeof(float));
  MatrixXd Ac(Na,Na);
  MatrixXd unshiftedAc(Na,Na);
  float m = mean(list_data,nx*ny);
  unsigned int hNa;

  for(unsigned int r=0;r<nx*ny;r++) list_data_0[r]=list_data[r]-m;

	if(Na<nx && Na<ny)
	{
	  do_fft_real(fft_list_data, list_data_0, nx, ny);
          pointwise_complex_squared_mod(squarred_fft_list_data , fft_list_data, nx*ny );
          do_ifft_real(squarred_fft_list_data, ifft_list_data, nx , ny );
	  hNa=(Na-1)/2;
	  for(unsigned int i=0;i<Na ;i++)
 	  {
	    for(unsigned int j=0;j<Na ;j++)
	    {
	      if((i<hNa) && (j<hNa)) Ac(j,i) = ifft_list_data[nx-hNa+i+(ny-hNa+j)*nx]/(nx*ny);
	      else if((i<hNa) && (j>hNa-1)) Ac(j,i) = ifft_list_data[nx-hNa+i+(j-hNa)*nx]/(nx*ny);
	      else if((i>hNa-1) && (j<hNa)) Ac(j,i) = ifft_list_data[i-hNa+(ny-hNa+j)*nx]/(nx*ny);
	      else Ac(j,i) = ifft_list_data[i-hNa+(j-hNa)*nx]/(nx*ny);
	    }
 	  }
	}

	/* free memory */
	fftwf_free(fft_list_data);
	fftwf_free(squarred_fft_list_data);
	free(ifft_list_data);
	free(list_data_0);

	return(Ac);
}

/**
* @brief Compute the cross correlation of arrays of float type.
*
* @param data_in input arrays
* @param N_data number of arrays
* @param N size of the arrays
* @return Cross correlation matrix of list_data (float matrix of size N_data*N_data)
*/
MatrixXf compute_cross_cor(float **list_data, size_t N_data, size_t N)
{
        float m=0;
	MatrixXf M(N_data, N);

	for(unsigned int i=0;i<N_data;i++)
	{
	  m = mean(list_data[i],N);
	  for(unsigned int j=0;j<N;j++)
	  {
	    M(i,j)= (float) (list_data[i][j]-m);
	  }
	}

	MatrixXf cross_cor=M*(M.transpose())/N;

	return(cross_cor);
}

/**
* @brief Compute the cross scale correlation of arrays of float type.
*
* @param list_data_x,list_data_y input arrays
* @param N_data_x,N_data_y number of arrays
* @param N size of the arrays
* @return Cross scale correlation matrix of list_data (float matrix)
*/
MatrixXf compute_cross_scale_cor(float **list_data_x, float **list_data_y,
				 size_t N_data_x, size_t N_data_y,
				 size_t N, int option)
{
        float mx=0;
	float my=0;
	MatrixXf X(N_data_x, N);
	MatrixXf Y(N_data_y, N);

	for(unsigned int i=0;i<N_data_x;i++)
	{
	  mx = mean(list_data_x[i],N);
	  for(unsigned int j=0;j<N;j++)
	  {
	    X(i,j)= (float) (list_data_x[i][j]-mx);
	  }
	}

	for(unsigned int i=0;i<N_data_y;i++)
	{
	 if(option ==1) my = mean(list_data_y[i],N);
	  for(unsigned int j=0;j<N;j++)
	  {
	    Y(i,j)= (float) (list_data_y[i][j]-my);
	  }
	}

	MatrixXf cross_scale_cor=X*(Y.transpose())/N;

	return(cross_scale_cor);
}

/**
* @brief Ajust the cross correlation of real arrays
*
* @param list_data list of modified arrays
* @param cross_cor_out desired cross correlation matrix
* @param N size of the array
* @return list_data with modified auto correlation
*/
void mod_cross_cor(float **list_data, const MatrixXf& cross_cor_out, size_t N)
{
  unsigned int N_data = cross_cor_out.rows();
  MatrixXf M(N_data, N);

  for(unsigned int i=0;i<N_data;i++)
  {
    for(unsigned int j=0;j<N;j++)
    {
      M(i,j)= (float) list_data[i][j];
    }
  }

  MatrixXf cross_cor_in = M*(M.transpose())/N;

  SelfAdjointEigenSolver<MatrixXf> eigensolver_in(cross_cor_in);
  SelfAdjointEigenSolver<MatrixXf> eigensolver_out(cross_cor_out);

  MatrixXf S_in=eigensolver_in.operatorInverseSqrt();

  MatrixXf S_out=eigensolver_out.operatorSqrt();

  MatrixXf K=S_out*S_in;

  M=K*M;

  for(unsigned int i=0;i<N_data;i++)
  {
    for(unsigned int j=0;j<N;j++)
    {
      list_data[i][j]=(float) M(i,j);
    }
  }
}

/**
* @brief Ajust the cross correlation and cross scale correlation of many real arrays
*
* @param list_data_x,list_data_y list of modified arrays
* @param cross_cor_out desired cross correlation matrix
* @param cross_scale_cor_out desired cross scale correlation matrix
* @param N size of the array
* @return mod an integer (to know if the input is modified or not)
* @note Note that list_data_x is modified
*/
int mod_cross_scale_cor(float **list_data_x, float **list_data_y, const MatrixXf& cross_cor_out, const MatrixXf& cross_scale_cor_out, size_t N)
{
  unsigned int N_data_x = cross_scale_cor_out.rows();
  unsigned int N_data_y = cross_scale_cor_out.cols();
  MatrixXf X(N_data_x, N);
  MatrixXf Y(N_data_y, N);
  int mod;

  for(unsigned int i=0;i<N_data_x;i++)
  {
    for(unsigned int j=0;j<N;j++)
    {
      X(i,j) = (float) list_data_x[i][j];
    }
  }

  for(unsigned int i=0;i<N_data_y;i++)
  {
    for(unsigned int j=0;j<N;j++)
    {
      Y(i,j) = (float) list_data_y[i][j];
    }
  }

  MatrixXf cross_cor_in = X*(X.transpose())/N;
  MatrixXf cross_scale_cor_in = X*(Y.transpose())/N;
  MatrixXf inv_cross_cor_in_prev_scale = (Y*(Y.transpose())/N).inverse();
  MatrixXf C_in = cross_cor_in-cross_scale_cor_in*inv_cross_cor_in_prev_scale*(cross_scale_cor_in.transpose());
  MatrixXf C_out = cross_cor_out-cross_scale_cor_out*inv_cross_cor_in_prev_scale*(cross_scale_cor_out.transpose());
  EigenSolver<MatrixXf> eigensolver_in(C_in);
  EigenSolver<MatrixXf> eigensolver_out(C_out);
  MatrixXcf D_in=eigensolver_in.eigenvalues().asDiagonal();
  MatrixXcf P_in=eigensolver_in.eigenvectors();
  MatrixXcf D_out=eigensolver_out.eigenvalues().asDiagonal();
  MatrixXcf P_out=eigensolver_out.eigenvectors();
  MatrixXcf isD_in=MatrixXcf::Zero(N_data_x,N_data_x);
  MatrixXcf sD_out=MatrixXcf::Zero(N_data_x,N_data_x);

  for(unsigned int i=0;i<N_data_x;i++)
  {
    if(fabs((float) D_in.real()(i,i)*D_in.real()(i,i)+D_in.imag()(i,i)*D_in.imag()(i,i))>0)
    {
      isD_in(i,i)=((complex<float>) 1)/sqrt((complex<float>) D_in(i,i));
    }
    sD_out(i,i)=sqrt((complex<float>) D_out(i,i));
  }

  MatrixXcf M=P_out*sD_out*P_out.transpose()*P_in*isD_in*P_in.transpose();
  MatrixXcf K=(cross_scale_cor_out.cast<complex<float> >()-M*cross_scale_cor_in.cast<complex<float> >())*inv_cross_cor_in_prev_scale.cast<complex<float> >();
  MatrixXcf Xnew = M*X.cast<complex<float> >()+K*Y.cast<complex<float> >();

  float testR=(Xnew.imag().cwiseAbs2().sum())/(Xnew.real().cwiseAbs2().sum());
  if(testR<10e-6)
  {
    mod=1;
    for(unsigned int i=0;i<N_data_x;i++)
    {
     for(unsigned int j=0;j<N;j++)
     {
      list_data_x[i][j]= (float ) Xnew.real()(i,j);
     }
    }
  }
  else mod=0;
  return(mod);
}

/**
* @brief Ajust the central auto correlation of an array
*
* @param data float array
* @param nx,ny size of the array
* @param auto_cor_out desired central auto correlation
* @return data with modified auto correlation
*/
void mod_auto_cor(float *data, const MatrixXd& auto_cor_out, size_t nx, size_t ny)
{
  unsigned int Na = auto_cor_out.rows();
  unsigned int hNa = (unsigned int) ((Na-1)/2);
  float m = mean(data,nx*ny);

  MatrixXd auto_cor_in = compute_auto_cor(data, nx, ny, 2*Na-1);

  int t = ((Na*Na+1)/2);

  MatrixXd A(t,t);
  VectorXd B(t);
  MatrixXd M(Na,Na);
  MatrixXd rM(Na,Na);

  for(unsigned int k=hNa;k<Na;k++)
  {
    if(k<Na-1)
    {
      for(unsigned int l=hNa;l<hNa+Na;l++)
      {
	M = auto_cor_in.block(k-hNa,l-hNa, Na, Na);
	
	for(unsigned int i=0;i<Na;i++)
	{
	  for(unsigned int j=0;j<Na;j++)
	  {
	    rM(i,j) = M(i,j) + M(Na-1-i,Na-1-j);
	  }
	}
	
	rM.resize(Na*Na,1);
	for(int u=0;u<t;u++){A((k-hNa)*Na+l-hNa,u)=rM(u,0);}
	rM.resize(Na,Na);
	A((k-hNa)*Na+l-hNa, t-1)/=2;	
	B((k-hNa)*Na+l-hNa) = auto_cor_out(k-hNa,l-hNa);
      }
    }
    else
    {
       for(unsigned int l=hNa;l<Na;l++)
       {
	M = auto_cor_in.block(k-hNa,l-hNa, Na, Na);
	
	for(unsigned int i=0;i<Na;i++)
	{
	  for(unsigned int j=0;j<Na;j++)
	  {
	    rM(i,j) = M(i,j) + M(Na-1-i,Na-1-j);
	  }
	}
	
	rM.resize(Na*Na,1);
	for( int u=0;u<t;u++){A((k-hNa)*Na+l-hNa,u)=rM(u,0);}
	rM.resize(Na,Na);
	A((k-hNa)*Na+l-hNa, t-1)/=2;
	B((k-hNa)*Na+l-hNa) = auto_cor_out(k-hNa,l-hNa);
      }
    }
  }

  VectorXd sol = A.colPivHouseholderQr().solve(B);
  MatrixXd fullsol(2*t-1,1);

  for(int i=0;i<t;i++)
  {
    fullsol(i,0)=sol(i);
    if(i<t-1){fullsol(t+i,0)=sol(t-i-2);}
  }

  fullsol.resize(Na,Na);
  MatrixXd hsquared0 = MatrixXd::Zero(ny,nx);
  hsquared0.block(ny/2-hNa,nx/2-hNa, Na, Na)=fullsol;
  MatrixXd shiftedhsquared0(ny,nx);
  shiftedhsquared0 << hsquared0.block(ny/2,nx/2,ny/2,nx/2),hsquared0.block(ny/2,0,ny/2,nx/2),
		      hsquared0.block(0,nx/2,ny/2,nx/2),hsquared0.block(0,0,ny/2,nx/2);

  float *h_in_fft = (float *) malloc(nx*ny*sizeof(float));
  for(unsigned int i=0;i<nx;i++) 
  {
    for(unsigned int j=0;j<ny;j++)
    {
      h_in_fft[i+j*nx]= (float) shiftedhsquared0(j,i);
    }
  }

  fftwf_complex *fft_h = (fftwf_complex *) fftwf_malloc(nx*ny*sizeof(fftwf_complex));
  do_fft_real(fft_h, h_in_fft, nx, ny);

  for(unsigned int i=0;i<nx;i++) 
  {
    for(unsigned int j=0;j<ny;j++)
    {
      fft_h[i+j*nx][0]=(float) sqrt(sqrt(fft_h[i+j*nx][0]*fft_h[i+j*nx][0]+fft_h[i+j*nx][1]*fft_h[i+j*nx][1]));
      fft_h[i+j*nx][1]=0;
      data[i+j*nx]-= m;
    }
  }

  fftwf_complex *fft_data = (fftwf_complex *) fftwf_malloc(nx*ny*sizeof(fftwf_complex));
  do_fft_real(fft_data, data, nx, ny);

  for(unsigned int i=0;i<nx*ny;i++)
  {
    fft_data[i][0]*=fft_h[i][0];
    fft_data[i][1]*=fft_h[i][0];
  }

  do_ifft_real(fft_data, data, nx, ny);
  for(unsigned int i=0;i<nx*ny;i++){data[i]+=m;}

  /* free memory */
  free(h_in_fft);
  fftwf_free(fft_data);
  fftwf_free(fft_h);
}