/*
 * Copyright (C) 2012, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright (C) 2012, Jonathan Vacher, ENS Cachan <jvacher@ens-cachan.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fftw3.h>
#include "eig3.h"
#include "parameters.h"

/**
* @file toolbox.c
* @brief Compute the bilinear zoom of an image.
*
* @version 1.0
* @author Thibaud BRIAND & Jonathan VACHER ; <thibaud.briand@ens-cachan.fr> ;
* <jvacher@ens-cachan.fr>
*
* @note Useful if you add the smooth component after using the periodic
* component option (see @ref periodic_component.c file and @ref ps function).
*/

/**
 * @brief Compute the DFT of a 2D real array.
 * @param data_in input array
 * @param nx,ny array size
 * @return fft_out complex array which corresponds to the DFT of the input
 */
void do_fft_real(fftwf_complex *fft_out, float *data_in, size_t nx , size_t ny)
{
  /* memory allocation */
  fftwf_plan plan_r2c;
  fftwf_complex *complex_data_in = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));

  /* Real --> complex */
  for(unsigned int i=0;i<nx*ny;i++)
  {
    complex_data_in[i][0]=data_in[i];
    complex_data_in[i][1]=0;
  }

  /* compute the DFT */
  plan_r2c = fftwf_plan_dft_2d((int) ny ,(int) nx , complex_data_in, fft_out , FFTW_FORWARD, FFTW_ESTIMATE);
  fftwf_execute(plan_r2c);

  /* free */
  fftwf_destroy_plan(plan_r2c);
  fftwf_free(complex_data_in);
  fftwf_cleanup();
}

/**
 * @brief Compute the DFT of a 2D real array.
 * @param data_in input array
 * @param nx,ny array size
 * @return fft_out complex array which corresponds to the DFT of the input
 */
void do_fft(fftwf_complex *fft_out, fftwf_complex *data_in, size_t nx , size_t ny)
{
  /* memory allocation */
  fftwf_plan plan_r2c;

  /* compute the DFT */
  plan_r2c = fftwf_plan_dft_2d((int) ny ,(int) nx , data_in, fft_out , FFTW_FORWARD, FFTW_ESTIMATE);
  fftwf_execute(plan_r2c);

  /* free */
  fftwf_destroy_plan(plan_r2c);
  fftwf_cleanup();
}

 /**
 * @brief Compute the iDFT of a 2D symmetric complex array.
 * @param fft_in input array
 * @param nx,ny array size
 * @return data_out real array which corresponds to the iDFT of the input
 */
void do_ifft_real(fftwf_complex *fft_in, float *data_out, size_t nx , size_t ny)
{
  /* memory allocation */
  fftwf_plan plan_c2r;
  fftwf_complex *complex_data_out = (fftwf_complex *) fftwf_malloc ( sizeof ( fftwf_complex ) * nx * ny );

  /* compute the iDFT */
  plan_c2r = fftwf_plan_dft_2d ( (int) ny, (int) nx , fft_in , complex_data_out , FFTW_BACKWARD, FFTW_ESTIMATE );
  fftwf_execute(plan_c2r);

  /* complex --> real + normalization */
  for(unsigned int i=0;i<nx*ny;i++)
    {
      data_out[i]=complex_data_out[i][0]/(nx*ny);
    }

  /* free */
  fftwf_free(complex_data_out);
  fftwf_destroy_plan(plan_c2r);
  fftwf_cleanup();
}

 /**
 * @brief Compute the iDFT of a 2D symmetric complex array.
 * @param fft_in input array
 * @param nx,ny array size
 * @return data_out real array which corresponds to the iDFT of the input
 */
void do_ifft(fftwf_complex *fft_in, fftwf_complex *data_out, size_t nx , size_t ny)
{
  /* memory allocation */
  fftwf_plan plan_c2r;

  /* compute the iDFT */
  plan_c2r = fftwf_plan_dft_2d ( (int) ny, (int) nx , fft_in , data_out , FFTW_BACKWARD, FFTW_ESTIMATE );
  fftwf_execute(plan_c2r);

   /* normalization */
  for(unsigned int i=0;i<nx*ny;i++)
    {
      data_out[i][0]/=(nx*ny);
      data_out[i][1]/=(nx*ny);
    }

  /* free */
  fftwf_destroy_plan(plan_c2r);
  fftwf_cleanup();
}

/**
* @brief Compute the change of basis according to the principal component analysis method.
*
* @param data_in_rgb input color array
* @param nx,ny array size
* @return M_eigen_vectors the 3x3 change of basis matrix, data_in_rgb_mean the mean of each color component
*/
void pca( float M_eigen_vectors[3][3], float data_in_rgb_mean[3],
		 float **data_in_rgb, size_t nx, size_t ny)
{
  float C[3][3];
  float d[3];

  /* Compute the mean of the input */
     for(int i=0;i<3;i++)    data_in_rgb_mean[i] = mean( data_in_rgb[i] , nx*ny );

  /* Compute the covariance matrix C*/
     for(int i=0;i<3;i++)
     {
       for(int j=0;j<i+1;j++)
       {
	 C[i][j]=0;
         for(unsigned int k=0; k<(int) nx*ny; k++) C[i][j] +=
	   (data_in_rgb[i][k] - data_in_rgb_mean[i])*(data_in_rgb[j][k] - data_in_rgb_mean[j]);
         C[j][i] = C[i][j]; /* the covariance matrix is symmetric */
       }
     }

  /* Compute the change of basis matrix */
     eigen_decomposition( C, M_eigen_vectors, d);
}